"use strict";
/*!
 * Copyright (c) 2020 Ron Buckton (rbuckton@chronicles.org)
 *
 * This file is licensed to you under the terms of the MIT License, found in the LICENSE file
 * in the root of this repository or package.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.GrammarkdownEmitter = void 0;
const emitter_1 = require("./emitter");
const tokens_1 = require("../tokens");
/** {@docCategory Emit} */
class GrammarkdownEmitter extends emitter_1.Emitter {
    constructor() {
        super(...arguments);
        this.extension = ".grammar";
    }
    emitProduction(node) {
        var _a, _b;
        this.emitIdentifier(node.name);
        this.emitNode(node.parameterList);
        this.writer.write(" ");
        this.emitTokenKind((_b = (_a = node.colonToken) === null || _a === void 0 ? void 0 : _a.kind) !== null && _b !== void 0 ? _b : tokens_1.SyntaxKind.ColonToken);
        this.emitNode(node.body);
    }
    emitParameterList(node) {
        this.writer.write("[");
        if (node.elements) {
            for (let i = 0; i < node.elements.length; ++i) {
                if (i > 0) {
                    this.writer.write(`, `);
                }
                this.emitNode(node.elements[i]);
            }
        }
        this.writer.write(`]`);
    }
    emitParameter(node) {
        this.emitIdentifier(node.name);
    }
    emitOneOfList(node) {
        var _a;
        this.writer.write(" ");
        this.emitTokenKind(tokens_1.SyntaxKind.OneKeyword);
        this.writer.write(" ");
        this.emitTokenKind(tokens_1.SyntaxKind.OfKeyword);
        this.writer.writeln();
        this.writer.indent();
        if ((_a = node.terminals) === null || _a === void 0 ? void 0 : _a.length) {
            for (const child of node.terminals) {
                this.emitNode(child);
            }
        }
        this.writer.dedent();
    }
    emitRightHandSideList(node) {
        this.writer.indent();
        if (node.elements) {
            for (const rhs of node.elements) {
                this.writer.writeln();
                this.emitNode(rhs);
            }
        }
        this.writer.dedent();
    }
    emitRightHandSide(node) {
        this.emitChildren(node);
    }
    emitSymbolSpan(node) {
        this.emitNode(node.symbol);
        if (node.next) {
            this.writer.write(" ");
            this.emitNode(node.next);
        }
    }
    emitPlaceholder(node) {
        this.emitTokenKind(tokens_1.SyntaxKind.AtToken);
    }
    emitTerminal(node) {
        var _a;
        if (node.text === "`") {
            this.writer.write("```");
        }
        else {
            this.writer.write("`");
            this.writer.write((_a = node.text) !== null && _a !== void 0 ? _a : "");
            this.writer.write("`");
        }
        this.emitNode(node.questionToken);
    }
    emitNonterminal(node) {
        this.emitChildren(node);
    }
    emitArgumentList(node) {
        this.writer.write("[");
        if (node.elements) {
            for (let i = 0; i < node.elements.length; ++i) {
                if (i > 0) {
                    this.writer.write(`, `);
                }
                this.emitNode(node.elements[i]);
            }
        }
        this.writer.write(`]`);
    }
    emitArgument(node) {
        this.emitToken(node.operatorToken);
        this.emitNode(node.name);
    }
    emitUnicodeCharacterRange(node) {
        this.emitTextContent(node.left);
        this.writer.write(` through `);
        this.emitTextContent(node.right);
    }
    emitUnicodeCharacterLiteral(node) {
        this.emitTextContent(node);
        this.emitNode(node.questionToken);
    }
    emitProse(node) {
        this.writer.write("> ");
        node.fragments && this.emitNodes(node.fragments);
    }
    emitEmptyAssertion(node) {
        this.writer.write("[empty]");
    }
    emitSymbolSet(node) {
        this.writer.write(`{`);
        if (node.elements) {
            for (let i = 0; i < node.elements.length; ++i) {
                if (i > 0) {
                    this.writer.write(`,`);
                }
                this.writer.write(` `);
                this.emitNode(node.elements[i]);
            }
        }
        this.writer.write(` }`);
    }
    emitLookaheadAssertion(node) {
        this.writer.write(`[`);
        if (node.operatorToken) {
            switch (node.operatorToken.kind) {
                case tokens_1.SyntaxKind.ExclamationEqualsToken:
                case tokens_1.SyntaxKind.NotEqualToToken:
                    this.writer.write(`lookahead ≠ `);
                    break;
                case tokens_1.SyntaxKind.EqualsToken:
                case tokens_1.SyntaxKind.EqualsEqualsToken:
                    this.writer.write(`lookahead = `);
                    break;
                case tokens_1.SyntaxKind.ElementOfToken:
                case tokens_1.SyntaxKind.LessThanMinusToken:
                    this.writer.write(`lookahead ∈ `);
                    break;
                case tokens_1.SyntaxKind.NotAnElementOfToken:
                case tokens_1.SyntaxKind.LessThanExclamationToken:
                    this.writer.write(`lookahead ∉ `);
                    break;
            }
        }
        this.emitNode(node.lookahead);
        this.writer.write(`]`);
    }
    emitLexicalGoalAssertion(node) {
        this.writer.write(`[lexical goal `);
        this.emitNode(node.symbol);
        this.writer.write(`]`);
    }
    emitNoSymbolHereAssertion(node) {
        this.writer.write(`[no `);
        if (node.symbols) {
            for (let i = 0; i < node.symbols.length; ++i) {
                if (i > 0) {
                    this.writer.write(` or `);
                }
                this.emitNode(node.symbols[i]);
            }
        }
        this.writer.write(` here]`);
    }
    emitConstraints(node) {
        this.writer.write("[");
        if (node.elements) {
            for (let i = 0; i < node.elements.length; ++i) {
                if (i > 0) {
                    this.writer.write(`, `);
                }
                this.emitNode(node.elements[i]);
            }
        }
        this.writer.write("]");
    }
    emitProseAssertion(node) {
        this.writer.write(`[>`);
        if (node.fragments) {
            for (const fragment of node.fragments) {
                this.emitNode(fragment);
            }
        }
        this.writer.write(`]`);
    }
    emitButNotSymbol(node) {
        this.emitNode(node.left);
        this.writer.write(` but not `);
        this.emitNode(node.right);
    }
    emitOneOfSymbol(node) {
        this.writer.write(`one of `);
        if (node.symbols) {
            for (let i = 0; i < node.symbols.length; ++i) {
                if (i > 0) {
                    this.writer.write(` or `);
                }
                this.emitNode(node.symbols[i]);
            }
        }
    }
    emitTextContent(node) {
        if (node === null || node === void 0 ? void 0 : node.text) {
            this.writer.write(node.text);
        }
    }
}
exports.GrammarkdownEmitter = GrammarkdownEmitter;
//# sourceMappingURL=grammarkdown.js.map