"use strict";
/*!
 * Copyright (c) 2020 Ron Buckton (rbuckton@chronicles.org)
 *
 * This file is licensed to you under the terms of the MIT License, found in the LICENSE file
 * in the root of this repository or package.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.NodeNavigator = void 0;
const tokens_1 = require("./tokens");
/**
 * Navigates the syntax-tree of a {@link SourceFile}.
 * {@docCategory Compiler}
 *
 * @remarks
 * Nodes in Grammarkdown's syntax tree are immutable and do not maintain pointers to their parents.
 * This can make traversing through a document somewhat difficult. The NodeNavigator class is intended
 * to improve this process by providing an API that can traverse a syntax tree starting from the root.
 *
 * A NodeNavigator focuses on a specific {@link Node} within a syntax tree, and maintains the
 * path to that node from the root. Various methods on the navigator move the focus, allowing you to
 * navigate to any other node within the syntax tree.
 */
class NodeNavigator {
    constructor(sourceFileOrNavigator) {
        this._copyOnNavigate = false;
        if (sourceFileOrNavigator instanceof NodeNavigator) {
            const navigator = sourceFileOrNavigator;
            this._sourceFile = navigator._sourceFile;
            this._edgeStack = navigator._edgeStack.slice();
            this._arrayStack = navigator._arrayStack.slice();
            this._offsetStack = navigator._offsetStack.slice();
            this._nodeStack = navigator._nodeStack.slice();
            this._currentDepth = this._nodeStack.length - 1;
        }
        else {
            this._sourceFile = sourceFileOrNavigator;
            this._edgeStack = [-1];
            this._arrayStack = [undefined];
            this._offsetStack = [0];
            this._nodeStack = [this._sourceFile];
            this._currentDepth = 0;
        }
        this._afterNavigate();
    }
    /**
     * Creates a copy of this {@link NodeNavigator} at the same position.
     */
    clone() {
        return new NodeNavigator(this);
    }
    /**
     * Gets the root {@link SourceFile} node for this navigator.
     */
    getRoot() {
        return this._sourceFile;
    }
    /**
     * Gets the parent {@link Node} of the {@link Node} the navigator is currently focused on.
     */
    getParent() {
        return this._parentNode;
    }
    /**
     * Gets the {@link Node} the navigator is currently focused on.
     */
    getNode() {
        return this._currentNode;
    }
    /**
     * If the {@link Node} the navigator is currently focused on is a {@link TextContentNode}, returns the `text` of the node;
     * Otherwise, returns `undefined`.
     */
    getTextContent() {
        if (tokens_1.isTextContentKind(this.getKind())) {
            return this._currentNode.text;
        }
    }
    /**
     * Gets the {@link SyntaxKind} of the {@link Node} the navigator is currently focused on.
     */
    getKind() {
        return this._currentNode.kind;
    }
    /**
     * Gets the string representation of the {@link SyntaxKind} of the {@link Node} the navigator is currently focused on.
     */
    getKindString() {
        return tokens_1.tokenToString(this.getKind());
    }
    /**
     * Gets the name of the property on the parent {@link Node} the navigator is currently focused on.
     */
    getName() {
        var _a;
        return (_a = this._parentNode) === null || _a === void 0 ? void 0 : _a.edgeName(this._currentEdge);
    }
    /**
     * Gets the containing node array of {@link Node} the navigator is currently focused on.
     */
    getArray() {
        return this._currentArray;
    }
    /**
     * Gets the ordinal offset within the containing node array of {@link Node} the navigator is currently focused on.
     */
    getOffset() {
        return this._currentOffset;
    }
    /**
     * Gets the current depth within the syntax-tree of the current focus of the navigator.
     */
    getDepth() {
        return this._currentDepth;
    }
    /**
     * Returns a value indicating whether the focus of the navigator points to a {@link Node} in an array.
     */
    isArray() {
        return this._currentArray !== undefined;
    }
    /**
     * Returns a value indicating whether the focus of the navigator points to either a {@link Token}, {@link TextContentNode}, or {@link InvalidSymbol}.
     */
    isToken() {
        const kind = this.getKind();
        return tokens_1.isTokenKind(kind)
            || tokens_1.isTextContentKind(kind)
            || kind === tokens_1.SyntaxKind.InvalidSymbol;
    }
    hasChildren(predicateOrKind) {
        if (this._hasAnyChildren === false || this._currentNode.edgeCount === 0) {
            return false;
        }
        if (predicateOrKind === undefined && this._hasAnyChildren !== undefined) {
            return this._hasAnyChildren;
        }
        for (let nextEdge = 0; nextEdge < this._currentNode.edgeCount; nextEdge++) {
            const next = this._currentNode.edgeValue(nextEdge);
            if (isNodeArray(next)) {
                for (let nextOffset = 0; nextOffset < next.length; nextOffset++) {
                    const nextNode = next[nextOffset];
                    if (nextNode && matchPredicateOrKind(nextNode, predicateOrKind)) {
                        return this._hasAnyChildren = true;
                    }
                }
            }
            else if (next && matchPredicateOrKind(next, predicateOrKind)) {
                return this._hasAnyChildren = true;
            }
        }
        if (!predicateOrKind) {
            this._hasAnyChildren = false;
        }
        return false;
    }
    hasAncestor(predicateOrKind) {
        for (let nextDepth = this._currentDepth - 1; nextDepth >= 0; nextDepth--) {
            const nextNode = this._nodeStack[nextDepth];
            if (matchPredicateOrKind(nextNode, predicateOrKind)) {
                return true;
            }
        }
        return false;
    }
    isMatch(predicateOrKind) {
        return matchPredicateOrKind(this._currentNode, predicateOrKind);
    }
    /**
     * Determines whether this navigator is focused on the same location within the tree as another navigator.
     * @param other The other navigator.
     * @returns `true` if both navigators are focused on the same location within the tree; otherwise, `false`.
     */
    isSamePosition(other) {
        if (this === other) {
            return true;
        }
        return this._sourceFile === other._sourceFile
            && this._currentDepth === other._currentDepth
            && this._currentEdge === other._currentEdge
            && this._currentArray === other._currentArray
            && this._currentOffset === other._currentOffset
            && this._currentNode === other._currentNode;
    }
    /**
     * Moves the focus of the navigator to the {@link Node} that contains the provided [Position](xref:grammarkdown!Position:interface).
     * @param position The [Position](xref:grammarkdown!Position:interface) at which to focus the navigator.
     * @param outermost When `true`, moves to the outermost node containing the provided position.
     * When `false` or not specified, moves to the innermost node containing the provided position.
     * @returns `true` if the navigator's focus changed; otherwise, `false`.
     */
    moveToPosition(position, outermost) {
        const pos = this._sourceFile.lineMap.offsetAt(position);
        const currentDepth = this._currentDepth;
        const edgeStack = this._edgeStack;
        const arrayStack = this._arrayStack;
        const offsetStack = this._offsetStack;
        const nodeStack = this._nodeStack;
        const hasChild = this._hasAnyChildren;
        this._copyOnNavigate = true;
        this.moveToRoot();
        if (this._moveToPositionWorker(pos, outermost)) {
            return true;
        }
        this._currentDepth = currentDepth;
        this._edgeStack = edgeStack;
        this._arrayStack = arrayStack;
        this._offsetStack = offsetStack;
        this._nodeStack = nodeStack;
        this._hasAnyChildren = hasChild;
        this._afterNavigate();
        return false;
    }
    _moveToPositionWorker(pos, outermost) {
        if (pos >= this._currentNode.pos && pos < this._currentNode.end) {
            if (!outermost) {
                if (this.moveToFirstChild()) {
                    do {
                        if (this._moveToPositionWorker(pos)) {
                            return true;
                        }
                    } while (this.moveToNextSibling());
                    this.moveToParent();
                }
            }
            return true;
        }
        return false;
    }
    /**
     * Moves the focus of this navigator to the same position within the syntax tree as another navigator.
     * @param other The other navigator.
     * @returns `true` if the navigator's focus changed; otherwise, `false`.
     */
    moveTo(other) {
        if (this === other) {
            return true;
        }
        if (this._sourceFile !== other._sourceFile) {
            return false;
        }
        this._currentDepth = other._currentDepth;
        this._edgeStack = other._edgeStack.slice();
        this._arrayStack = other._arrayStack.slice();
        this._offsetStack = other._offsetStack.slice();
        this._nodeStack = other._nodeStack.slice();
        this._afterNavigate();
        return true;
    }
    /**
     * Moves the focus of the navigator to the root of the syntax tree.
     * @returns `true` if the navigator's focus changed; otherwise, `false`.
     */
    moveToRoot() {
        if (this._currentDepth > 0) {
            this._beforeNavigate();
            this._reset();
            this._afterNavigate();
        }
        return true;
    }
    moveToParent(predicateOrKind) {
        if (this._currentDepth > 0) {
            if (matchPredicateOrKind(this._parentNode, predicateOrKind)) {
                this._beforeNavigate();
                this._popEdge();
                this._afterNavigate();
                return true;
            }
        }
        return false;
    }
    moveToAncestorOrSelf(predicateOrKind) {
        return this.isMatch(predicateOrKind)
            || this.moveToAncestor(predicateOrKind);
    }
    moveToAncestor(predicateOrKind) {
        for (let nextDepth = this._currentDepth - 1; nextDepth >= 0; nextDepth--) {
            const nextNode = this._nodeStack[nextDepth];
            if (matchPredicateOrKind(nextNode, predicateOrKind)) {
                this._beforeNavigate();
                while (this._currentDepth !== nextDepth) {
                    this._popEdge();
                }
                this._afterNavigate();
                return true;
            }
        }
        return false;
    }
    /**
     * Moves the focus of the navigator to the parent of the focused {@link Node} if that parent is a {@link SourceElement}.
     * @returns `true` if the navigator's focus changed; otherwise, `false`.
     */
    moveToSourceElement() {
        return this.moveToParent(matchSourceElement);
    }
    /**
     * Moves the focus of the navigator to the parent of the focused {@link Node} if that parent is either a {@link Parameter} or a {@link Production}.
     * @returns `true` if the navigator's focus changed; otherwise, `false`.
     */
    moveToDeclaration() {
        return this.moveToParent(tokens_1.SyntaxKind.Parameter)
            || this.moveToParent(tokens_1.SyntaxKind.Production);
    }
    /**
     * Moves the focus of the navigator to the nearest {@link Identifier}.
     * @returns `true` if the current node is an {@link Identifier} or the navigator's focus changed; otherwise, `false`.
     * @remarks
     * The "nearest {@link Identifier}" is determined using the following rules:
     * <ul>
     * <li>If the focus or its nearest ancestor is a {@link Parameter}, move to the `name` of the {@link Parameter}.</li>
     * <li>If the focus or its nearest ancestor is an {@link Argument}, move to the `name` of the {@link Argument}.</li>
     * <li>If the focus or its nearest ancestor is a {@link Nonterminal}, move to the `name` of the {@link Nonterminal}.</li>
     * <li>If the focus or its nearest ancestor is a {@link LexicalGoalAssertion}, move to the `symbol` of the of the {@link LexicalGoalAssertion}.</li>
     * <li>If the focus or its nearest ancestor is a {@link Define}, move to the `key` of the {@link Define}.</li>
     * <li>If the focus or its nearest ancestor is a {@link Constraints}, move to the `name` of the of the first {@link Argument} of the {@link Constraints}.</li>
     * <li>If the focus is not within the `body` of a {@link Production} and the focus or its nearest ancestor is a {@link Production}, move to the `name` of the {@link Production}.</li>
     * </ul>
     */
    moveToName() {
        if (this.getKind() === tokens_1.SyntaxKind.Identifier) {
            return true;
        }
        else {
            const navigator = this.clone();
            if (navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.Parameter)
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.Argument)
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.Nonterminal)
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.LexicalGoalAssertion)
                && navigator.moveToFirstChild("symbol")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.Constraints)
                && navigator.moveToFirstChild("elements")
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (!navigator.hasAncestor(matchProductionBody)
                && navigator.moveToAncestorOrSelf(tokens_1.SyntaxKind.Production)
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
            navigator.moveTo(this);
            if (navigator.moveToParent(tokens_1.SyntaxKind.InvalidAssertion)
                && navigator.moveToParent(tokens_1.SyntaxKind.SymbolSpan)
                && navigator.moveToPreviousSibling(tokens_1.SyntaxKind.Nonterminal)
                && navigator.moveToFirstChild("name")) {
                return this.moveTo(navigator);
            }
        }
        return false;
    }
    moveToFirstChild(predicateOrNameOrKind) {
        return this._moveToChild(Navigation.first, Navigation.next, predicateOrNameOrKind);
    }
    moveToLastChild(predicateOrNameOrKind) {
        return this._moveToChild(Navigation.last, Navigation.previous, predicateOrNameOrKind);
    }
    moveToFirstElement(predicateOrKind) {
        return this._moveToElement(Navigation.first, Navigation.next, this._currentEdge, this._currentArray, this._currentOffset, predicateOrKind);
    }
    moveToPreviousElement(predicateOrKind) {
        return this._moveToElement(Navigation.previous, Navigation.previous, this._currentEdge, this._currentArray, this._currentOffset, predicateOrKind);
    }
    moveToNextElement(predicateOrKind) {
        return this._moveToElement(Navigation.next, Navigation.next, this._currentEdge, this._currentArray, this._currentOffset, predicateOrKind);
    }
    moveToLastElement(predicateOrKind) {
        return this._moveToElement(Navigation.last, Navigation.previous, this._currentEdge, this._currentArray, this._currentOffset, predicateOrKind);
    }
    moveToFirstSibling(predicateOrNameOrKind) {
        return this._moveToSibling(Navigation.first, undefined, Navigation.first, Navigation.next, this._parentNode, predicateOrNameOrKind);
    }
    moveToPreviousSibling(predicateOrNameOrKind) {
        return this._moveToSibling(Navigation.previous, Navigation.previous, Navigation.last, Navigation.previous, this._parentNode, predicateOrNameOrKind);
    }
    moveToNextSibling(predicateOrNameOrKind) {
        return this._moveToSibling(Navigation.next, Navigation.next, Navigation.first, Navigation.next, this._parentNode, predicateOrNameOrKind);
    }
    moveToLastSibling(predicateOrNameOrKind) {
        return this._moveToSibling(Navigation.last, undefined, Navigation.last, Navigation.previous, this._parentNode, predicateOrNameOrKind);
    }
    moveToFirstToken(predicateOrKind) {
        if (this.isToken())
            return true;
        const navigator = this.clone();
        while (!navigator.isToken()) {
            if (!navigator.moveToFirstChild()) {
                return false;
            }
        }
        return navigator.isMatch(predicateOrKind) && this.moveTo(navigator);
    }
    moveToLastToken(predicateOrKind) {
        if (this.isToken())
            return true;
        const navigator = this.clone();
        while (!navigator.isToken()) {
            if (!navigator.moveToLastChild()) {
                return false;
            }
        }
        return navigator.isMatch(predicateOrKind) && this.moveTo(navigator);
    }
    moveToNextToken(predicateOrKind) {
        const navigator = this.clone();
        while (!navigator.moveToNextSibling()) {
            if (!navigator.moveToParent()) {
                return false;
            }
        }
        if (!navigator.moveToFirstToken()) {
            return false;
        }
        return navigator.isMatch(predicateOrKind) && this.moveTo(navigator);
    }
    moveToPreviousToken(predicateOrKind) {
        const navigator = this.clone();
        while (!navigator.moveToPreviousSibling()) {
            if (!navigator.moveToParent()) {
                return false;
            }
        }
        if (!navigator.moveToLastToken()) {
            return false;
        }
        return navigator.isMatch(predicateOrKind) && this.moveTo(navigator);
    }
    _beforeNavigate() {
        if (this._copyOnNavigate) {
            this._edgeStack = this._edgeStack.slice();
            this._arrayStack = this._arrayStack.slice();
            this._offsetStack = this._offsetStack.slice();
            this._nodeStack = this._nodeStack.slice();
            this._copyOnNavigate = false;
        }
    }
    _afterNavigate() {
        this._currentEdge = this._edgeStack[this._currentDepth];
        this._currentArray = this._arrayStack[this._currentDepth];
        this._currentOffset = this._offsetStack[this._currentDepth];
        this._currentNode = this._nodeStack[this._currentDepth];
        this._parentNode = this._currentDepth > 0 ? this._nodeStack[this._currentDepth - 1] : undefined;
        this._copyOnNavigate = false;
    }
    _pushEdge() {
        this._edgeStack.push(undefined);
        this._arrayStack.push(undefined);
        this._offsetStack.push(undefined);
        this._nodeStack.push(undefined);
        this._hasAnyChildren = undefined;
        this._currentDepth++;
    }
    _setEdge(edge, array, offset, node) {
        this._edgeStack[this._currentDepth] = edge;
        this._arrayStack[this._currentDepth] = array;
        this._offsetStack[this._currentDepth] = offset;
        this._nodeStack[this._currentDepth] = node;
        this._hasAnyChildren = undefined;
    }
    _popEdge() {
        this._currentDepth--;
        this._edgeStack.pop();
        this._arrayStack.pop();
        this._offsetStack.pop();
        this._nodeStack.pop();
        this._hasAnyChildren = this._currentNode !== undefined;
    }
    _moveToChild(initializer, seekDirection, predicateOrNameOrKind) {
        const name = typeof predicateOrNameOrKind === "string" ? predicateOrNameOrKind : undefined;
        const predicateOrKind = typeof predicateOrNameOrKind !== "string" ? predicateOrNameOrKind : undefined;
        const offset = this._currentEdge;
        const length = this._currentNode.edgeCount;
        for (let nextEdge = initializer(offset, length); bounded(nextEdge, length); nextEdge = seekDirection(nextEdge, length)) {
            if (!name || this._currentNode.edgeName(nextEdge) === name) {
                const next = this._currentNode.edgeValue(nextEdge);
                if (isNodeArray(next)) {
                    const length = next.length;
                    for (let nextOffset = initializer(0, length); bounded(nextOffset, length); nextOffset = seekDirection(nextOffset, length)) {
                        const nextNode = next[nextOffset];
                        if (nextNode && matchPredicateOrKind(nextNode, predicateOrKind)) {
                            this._beforeNavigate();
                            this._pushEdge();
                            this._setEdge(nextEdge, next, nextOffset, nextNode);
                            this._afterNavigate();
                            return true;
                        }
                    }
                }
                else if (next && matchPredicateOrKind(next, predicateOrKind)) {
                    this._beforeNavigate();
                    this._pushEdge();
                    this._setEdge(nextEdge, /*array*/ undefined, /*offset*/ undefined, next);
                    this._afterNavigate();
                    return true;
                }
            }
        }
        return false;
    }
    _moveToElement(currentArrayInitializer, seekDirection, currentEdge, currentArray, currentOffset, predicateOrKind) {
        if (!currentArray) {
            return false;
        }
        const offset = currentOffset;
        const length = currentArray.length;
        for (let nextOffset = currentArrayInitializer(offset, length); bounded(nextOffset, length); nextOffset = seekDirection(nextOffset, length)) {
            const nextNode = currentArray[nextOffset];
            if (nextNode && matchPredicateOrKind(nextNode, predicateOrKind)) {
                this._beforeNavigate();
                this._setEdge(currentEdge, currentArray, nextOffset, nextNode);
                this._afterNavigate();
                return true;
            }
        }
        return false;
    }
    _moveToSibling(currentEdgeInitializer, currentArrayInitializer, nextArrayInitializer, seekDirection, parentNode, predicateOrNameOrKind) {
        if (!parentNode) {
            return false;
        }
        const name = typeof predicateOrNameOrKind === "string" ? predicateOrNameOrKind : undefined;
        const predicateOrKind = typeof predicateOrNameOrKind !== "string" ? predicateOrNameOrKind : undefined;
        if (currentArrayInitializer && this._moveToElement(currentArrayInitializer, seekDirection, this._currentEdge, this._currentArray, this._currentOffset, predicateOrKind)) {
            return true;
        }
        const offset = this._currentEdge;
        const length = parentNode.edgeCount;
        for (let nextEdge = currentEdgeInitializer(offset, length); bounded(nextEdge, length); nextEdge = seekDirection(nextEdge, length)) {
            if (!name || parentNode.edgeName(nextEdge) === name) {
                const next = parentNode.edgeValue(nextEdge);
                if (isNodeArray(next)) {
                    if (this._moveToElement(nextArrayInitializer, seekDirection, nextEdge, next, 0, predicateOrKind)) {
                        return true;
                    }
                }
                else {
                    if (next && matchPredicateOrKind(next, predicateOrKind)) {
                        this._beforeNavigate();
                        this._setEdge(nextEdge, /*array*/ undefined, /*offset*/ undefined, next);
                        this._afterNavigate();
                        return true;
                    }
                }
            }
        }
        return false;
    }
    _reset() {
        this._currentDepth = 0;
        this._nodeStack.length = 1;
        this._edgeStack.length = 1;
        this._arrayStack.length = 1;
        this._offsetStack.length = 1;
    }
}
exports.NodeNavigator = NodeNavigator;
var Navigation;
(function (Navigation) {
    function first(offset, length) {
        return 0;
    }
    Navigation.first = first;
    function previous(offset, length) {
        return offset - 1;
    }
    Navigation.previous = previous;
    function same(offset, length) {
        return offset;
    }
    Navigation.same = same;
    function next(offset, length) {
        return offset + 1;
    }
    Navigation.next = next;
    function last(offset, length) {
        return length - 1;
    }
    Navigation.last = last;
})(Navigation || (Navigation = {}));
function bounded(offset, length) {
    return offset >= 0 && offset < length;
}
function matchProductionBody(node) {
    return node.kind === tokens_1.SyntaxKind.OneOfList
        || node.kind === tokens_1.SyntaxKind.RightHandSideList
        || node.kind === tokens_1.SyntaxKind.RightHandSide;
}
function matchSourceElement(node) {
    return node.kind === tokens_1.SyntaxKind.Import
        || node.kind === tokens_1.SyntaxKind.Production;
}
function isNodeArray(value) {
    return Array.isArray(value);
}
function matchPredicateOrKind(node, predicateOrKind) {
    return typeof predicateOrKind === "function" ? predicateOrKind(node) :
        typeof predicateOrKind === "number" ? node.kind === predicateOrKind :
            true;
}
//# sourceMappingURL=navigator.js.map